/*
 * Copyright (c) 2013, SRI International
 * All rights reserved.
 * Licensed under the The BSD 3-Clause License;
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 * 
 * http://opensource.org/licenses/BSD-3-Clause
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of the aic-expresso nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.sri.ai.brewer.parsingexpression.helper;


import java.util.List;

import com.google.common.annotations.Beta;
import com.google.common.base.Function;
import com.sri.ai.brewer.api.Writer;
import com.sri.ai.brewer.core.DefaultWriter;
import com.sri.ai.brewer.parsingexpression.core.Sequence;
import com.sri.ai.expresso.api.Expression;
import com.sri.ai.expresso.api.SyntaxTree;
import com.sri.ai.expresso.core.DefaultSymbol;
import com.sri.ai.expresso.helper.Expressions;
import com.sri.ai.util.Util;

/**
 * An associative sequence.
 * 
 * @author braz
 *
 */
@Beta
public class AssociativeSequence extends Sequence {
	private static final long serialVersionUID = 1L;
	
	public AssociativeSequence(Object... args) {
		super(args);
		label = DefaultSymbol.createSymbol("associative sequence");
	}

	@Override
	protected SyntaxTree postProcessParse(SyntaxTree parse) {
		if (parse.numberOfArguments() > 0) {
			return Expressions.associateWhenSureOperatorIsAssociative(parse).getSyntaxTree();
		}
		return parse;
	}

	@Override
	/**
	 * The arity of expressions parsed from 
	 * associative sequence parsing expressions is -1 ("variable arity").
	 */
	public int getArity() {
		return -1;
	}
	
	@Override
	public String toString(final Expression expression, final Writer writer) {
		StringBuffer result = new StringBuffer();
		String separator = " ";
		final SyntaxTree syntaxTree = DefaultWriter.getSyntaxTreeOrNullIfNull(expression);
		SyntaxTree rootTree = syntaxTree.getRootTree();
		String rootTreeString = writer.toString(rootTree);
		
		if (rootTree.numberOfImmediateSubTrees() > 0) {
			rootTreeString = "(" + rootTreeString + ")";
		}
		
		if (syntaxTree.numberOfImmediateSubTrees() < 2) {
			List<String> subTreesRepresentation =
				Util.mapIntoList(
						syntaxTree.getImmediateSubTrees(),
						new Function<SyntaxTree, String>() { 
							@Override
							public String apply(SyntaxTree subTree) {
								return DefaultWriter.subTreeRepresentation(
									syntaxTree,
									subTree,
									writer,
									false /* no need for parentheses */);
						}});
			return rootTreeString + "(" + Util.join("," + separator, subTreesRepresentation) + ")";
		}
		
		int subTreeIndex = 0;
		for (SyntaxTree subTree : syntaxTree.getImmediateSubTrees()) {
			if (subTreeIndex != 0) {
				result.append(separator + rootTreeString + separator);
			}
			result.append(DefaultWriter.subTreeRepresentation(syntaxTree, subTree, writer, false));
			subTreeIndex++;
		}
		return result.toString();
	}
}
