/*
 * Copyright (c) 2013, SRI International
 * All rights reserved.
 * Licensed under the The BSD 3-Clause License;
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 * 
 * http://opensource.org/licenses/BSD-3-Clause
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of the aic-expresso nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.sri.ai.grinder.library.set.tuple;

import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import com.google.common.annotations.Beta;
import com.sri.ai.expresso.api.Expression;
import com.sri.ai.expresso.api.ExpressionAndContext;
import com.sri.ai.expresso.core.DefaultExpressionAndContext;
import com.sri.ai.expresso.core.DefaultCompoundSyntaxTree;
import com.sri.ai.expresso.core.DefaultSymbol;
import com.sri.ai.expresso.helper.ExpressionKnowledgeModule;
import com.sri.ai.expresso.helper.Expressions;
import com.sri.ai.grinder.api.NoOpRewriter;
import com.sri.ai.grinder.api.RewritingProcess;
import com.sri.ai.grinder.core.AbstractRewriter;
import com.sri.ai.grinder.helper.BasePathIterator;
import com.sri.ai.grinder.library.function.InjectiveModule;
import com.sri.ai.grinder.library.function.MutuallyExclusiveCoDomainsModule;
import com.sri.ai.util.Util;
import com.sri.ai.util.collect.PairIterator;

/**
 * API for tuples. Because parsing tuples with single elements is ambiguous with
 * parenthesized expressions, function applications of "tuple" are also treated
 * as tuples.
 * 
 * @author braz
 */
@Beta
public class Tuple extends AbstractRewriter
implements
NoOpRewriter,
ExpressionKnowledgeModule.Provider,
InjectiveModule.Provider,
MutuallyExclusiveCoDomainsModule.Provider {

	public static final String TUPLE_LABEL = "( . )";
	//
	private static final List<Integer> _pathZero = Collections.unmodifiableList(Arrays.asList(new Integer(0)));

	@Override
	public Expression rewriteAfterBookkeeping(Expression expression, RewritingProcess process) {
		// Note: is a NoOpRewriter
		return expression;
	}
	
	@SuppressWarnings("unchecked")
	public static Expression make(Object... elements) {
		List list = null;
		if (elements.length == 1 && elements[0] instanceof List) {
			list = (List) elements[0];
		}
		else {
			list = Arrays.asList(elements);
		}
		DefaultCompoundSyntaxTree result = new DefaultCompoundSyntaxTree(TUPLE_LABEL, Expressions.makeKleeneListIfNeeded(list));
		return result;
	}

	public static boolean isTuple(Expression expression) {
		boolean result =
				expression.getSyntaxTree().getLabel().equals(TUPLE_LABEL)
				||
				expression.hasFunctor("tuple");
		return result;
	}
	
	public static List<Expression> getElements(Expression expression) {
		List<Expression> elements;
		if (expression.hasFunctor("tuple")) {
			elements = expression.getArguments();
		}
		else {
			elements = Expressions.ensureListFromKleeneList(expression.getSyntaxTree().getSubTree(0));
		}
		return elements;
	}
	
	public static int size(Expression expression) {
		List<Expression> elements = getElements(expression);
		int result = elements.size();
		return result;
	}
	
	public static Expression get(Expression expression, int index) {
		return getElements(expression).get(index);
	}

	@Override
	public boolean haveMutuallyExclusiveCoDomains(Expression expression1, Expression expression2, RewritingProcess process) {
		boolean result = isTuple(expression1) &&
				isTuple(expression2) &&
				size(expression1) != size(expression2);
		return result;
	}

	@Override
	public Object getInjectiveFunctionToken(Expression expression, RewritingProcess process) {
		if ( ! isTuple(expression)) {
			return null;
		}
		Expression token = DefaultSymbol.createSymbol("tuple/" + size(expression));
		return token;
	}

	@Override
	public void rewritingProcessInitiated(RewritingProcess process) {
		ExpressionKnowledgeModule knowledgeBasedExpressionModule =
				(ExpressionKnowledgeModule) process.findModule(ExpressionKnowledgeModule.class);
		if (knowledgeBasedExpressionModule != null) {
			knowledgeBasedExpressionModule.register(this);
		}

		InjectiveModule injectiveModuleModule = (InjectiveModule) process.findModule(InjectiveModule.class);
		if (injectiveModuleModule != null) {
			injectiveModuleModule.register(this);
		}

		MutuallyExclusiveCoDomainsModule mutuallyExclusiveCoDomainsModule = (MutuallyExclusiveCoDomainsModule) process.findModule(MutuallyExclusiveCoDomainsModule.class);
		if (mutuallyExclusiveCoDomainsModule != null) {
			mutuallyExclusiveCoDomainsModule.register(this);
		}
	}

	@Override
	public Object getSyntacticFormType(Expression expression, RewritingProcess process) {
		if (isTuple(expression)) {
			return "Tuple";
		}
		return null;
	}

	@Override
	public Iterator<ExpressionAndContext> getImmediateSubExpressionsAndContextsIterator(
			Expression expression, RewritingProcess process) {
		if (isTuple(expression)) {
			if (expression.hasFunctor("tuple")) {
				return Expressions.getSubExpressionsAndContextsIteratorFromImmediateSubTrees(expression, process);
			}
			else {
				List<Expression> elements = getElements(expression);
				List<Integer> basePath = _pathZero;
				if (size(expression) == 1) { // when there is a single element, there is no kleene list in the representation, and the single element is the only argument.
					return Util.iterator((ExpressionAndContext) new DefaultExpressionAndContext(elements.get(0), basePath));
				}

				PairIterator<Expression, List<Integer>> expressionAndPathPairsIterator = new PairIterator<Expression, List<Integer>>(
						elements.iterator(),
						new BasePathIterator(basePath, 0, size(expression)));
				Iterator<ExpressionAndContext> result = DefaultExpressionAndContext.makeExpressionAndContextIteratorFromPairs(expressionAndPathPairsIterator);
				return result;
			}
		}
		return null;
	}
}
